from dataclasses_jsonschema import JsonSchemaMixin, FieldEncoder, JsonDict
from typing import NewType
from datetime import datetime
from base64 import b64decode, b64encode

Binary = NewType("Binary", bytes)
DateTime = NewType("DateTime", datetime)


class BinaryField(FieldEncoder):
    """Base64 encoded bytes"""

    def to_wire(self, value: Binary) -> str:
        return b64encode(value).decode("utf-8")

    def to_python(self, value: str) -> bytes:
        return b64decode(value)

    @property
    def json_schema(self) -> JsonDict:
        return {
            "type": "string",
            "pattern": r"^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$",
        }


JsonSchemaMixin.register_field_encoders({Binary: BinaryField()})


class DateTimeField(FieldEncoder):
    """Wrapper around Python datetime.datetime object"""

    def to_wire(self, value: DateTime) -> float:
        if isinstance(value, float):
            return value
        return value.timestamp()

    def to_python(self, value: float) -> DateTime:
        return DateTime(datetime.fromtimestamp(value))

    @property
    def json_schema(self) -> JsonDict:
        return {"type": "number"}


JsonSchemaMixin.register_field_encoders({DateTime: DateTimeField()})



import redis
import fakeredis
from typing import Optional
from redorm.settings import REDORM_URL
from redorm.exceptions import DevelopmentServerUsedInProd


class RedormClient:
    pool: redis.ConnectionPool
    client: redis.Redis
    get_set_indirect_sha: Optional[str] = None
    get_key_indirect_sha: Optional[str] = None

    def __init__(self):
        self.server = None
        self.bind(REDORM_URL)

    def bind(self, url) -> bool:
        if url:
            self.client = redis.Redis.from_url(url, decode_responses=True)
            self.get_key_indirect_sha = self.client.register_script(
                "return {redis.call('get', redis.call('get', KEYS[1]))}"
            ).sha
            self.get_set_indirect_sha = self.client.register_script(
                """local l = {}
local keys = redis.call('smembers', KEYS[2])
for _,k in ipairs(keys) do
table.insert(l, redis.call('get', KEYS[1] .. k))
end
return l
"""
            ).sha
            return True
        else:
            # Lightweight fake redis
            self.client = fakeredis.FakeRedis(decode_responses=True)
            return False

    def init_app(self, app):
        url = app.config.get("REDORM_URL")
        if url:
            success = self.bind(url)
            if not success and app.config.get("FLASK_ENV") == "production" and not app.config.get("REDORM_INBUILT"):
                raise DevelopmentServerUsedInProd(
                    "Redorm was not provided a Redis URL and would fall back to the in-built Redis. "
                    "In the flask app config either set REDORM_URL to a valid redis connection string "
                    "or set REDORM_INBUILT to True (not recommended)"
                )


red = RedormClient()



from typing import List, Type, Optional, TypeVar, Union
from enum import Enum, auto
from redorm.model import RedormBase, all_models, IRelationship
from redorm.client import red

__all__ = [
    "RelationshipConfigEnum",
    "Relationship",
    "one_to_many",
    "one_to_one",
    "many_to_one",
    "many_to_many",
]


class RelationshipConfigEnum(Enum):
    ONE_TO_MANY = auto()
    ONE_TO_ONE = auto()
    MANY_TO_ONE = auto()
    MANY_TO_MANY = auto()


T = TypeVar("T", bound=RedormBase)
U = TypeVar("U", bound=RedormBase)


# noinspection PyProtectedMember
class Relationship(IRelationship):
    def __init__(
        self,
        foreign_type: Union[str, Type[U]],
        config: RelationshipConfigEnum = RelationshipConfigEnum.MANY_TO_MANY,
        backref: Optional[str] = None,
        lazy: bool = True,
    ):
        self.config = config
        if isinstance(foreign_type, str):
            self.foreign_type = None
            self.__foreign_type = foreign_type
        else:
            self.foreign_type = foreign_type
            self.__foreign_type = foreign_type.__name__
        self.fdel = None
        self.backref = backref
        self.to_many = config in {
            RelationshipConfigEnum.MANY_TO_MANY,
            RelationshipConfigEnum.ONE_TO_MANY,
        }
        self.many_to = config in {
            RelationshipConfigEnum.MANY_TO_MANY,
            RelationshipConfigEnum.MANY_TO_ONE,
        }
        self.relationship_name = None
        self.lazy = lazy
        self.__doc__ = None
        self.__owner = None
        self.cached_class = None
        self.cached_ref = None
        self.relationship_base = None

    def __set_name__(self, owner, name):
        self.__owner = owner
        self.relationship_name = name
        self.relationship_base = f"{owner.__name__}:relationship:{self.relationship_name}"
        owner._relationships = dict(**owner._relationships, **{name: self})

    def get_foreign_type(self) -> Type[U]:
        if self.foreign_type is None:
            self.foreign_type = all_models[self.__foreign_type]
            return self.foreign_type
        else:
            return self.foreign_type

    def __get__(self, instance: T, objtype=None):
        if instance is None:
            return self
        if self.cached_class is not None:
            return self.cached_class
        if not self.lazy:
            print("Cache miss!")
        relationship_path = f"{self.relationship_base}:{instance.id}"
        if self.to_many:
            if self.cached_ref is None:
                self.cached_ref = red.client.smembers(relationship_path)
            return self.get_foreign_type().get_bulk(self.cached_ref)
        else:
            if self.cached_ref is None:
                self.cached_ref = red.client.get(relationship_path)
            return self.get_foreign_type().get(self.cached_ref) if self.cached_ref is not None else None

    def __set__(
        self,
        instance: T,
        value: Union[None, str, U, List[Union[str, U]]],
    ):
        foreign_type = self.get_foreign_type()
        relationship_path = f"{self.relationship_base}:{instance.id}"
        if self.config in {
            RelationshipConfigEnum.MANY_TO_ONE,
            RelationshipConfigEnum.ONE_TO_ONE,
        }:
            related_id_new: Optional[str]
            if isinstance(value, RedormBase):
                related_id_new = value.id
            elif isinstance(value, str):
                related_id_new = value
            elif value is None:
                related_id_new = None
            else:
                raise ValueError("Expected new value of string or Model")
            if related_id_new is None:
                pipeline = red.client.pipeline()
                pipeline.get(relationship_path)
                pipeline.delete(relationship_path)
                result = pipeline.execute()
                related_id_old = result[0]
            else:
                related_id_old = red.client.getset(
                    relationship_path,
                    related_id_new,
                )
            if related_id_new == related_id_old or self.backref is None:
                return
            rel_new = f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_new}"
            rel_old = f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_old}"
            if self.config == RelationshipConfigEnum.MANY_TO_ONE:
                if related_id_old is None:
                    red.client.sadd(
                        rel_new,
                        instance.id,
                    )
                elif related_id_new is None:
                    red.client.srem(
                        rel_old,
                        instance.id,
                    )
                else:
                    red.client.smove(
                        rel_old,
                        rel_new,
                        instance.id,
                    )
            else:
                if related_id_old is None:
                    red.client.set(
                        rel_new,
                        instance.id,
                    )
                elif related_id_new is None:
                    red.client.delete(
                        rel_new,
                    )
                else:
                    red.client.rename(
                        rel_old,
                        rel_new,
                    )
        else:
            if (not isinstance(value, list)) and (not isinstance(value, set)):
                raise ValueError("Expected list or set for new relationships")
            old_related_ids = {r for r in red.client.smembers(relationship_path)}
            new_related_ids = {(r.id if isinstance(r, RedormBase) else r) for r in value}
            if len(old_related_ids.symmetric_difference(new_related_ids)) == 0:
                return
            pipeline = red.client.pipeline()
            pipeline.delete(relationship_path)
            if new_related_ids:
                pipeline.sadd(
                    relationship_path,
                    *new_related_ids,
                )
            if self.backref is None:
                pipeline.execute()
                return
            ids_to_remove = old_related_ids - new_related_ids
            ids_to_add = new_related_ids - old_related_ids
            reverse_path = f"{foreign_type.__name__}:relationship:{self.backref}"
            if self.config == RelationshipConfigEnum.MANY_TO_MANY:
                for idr in ids_to_remove:
                    pipeline.srem(
                        f"{reverse_path}:{idr}",
                        instance.id,
                    )
                for ida in ids_to_add:
                    pipeline.sadd(
                        f"{reverse_path}:{ida}",
                        instance.id,
                    )
            elif self.config == RelationshipConfigEnum.ONE_TO_MANY:
                for idr in ids_to_remove:
                    pipeline.delete(f"{reverse_path}:{idr}")
                for ida in ids_to_add:
                    pipeline.set(
                        f"{reverse_path}:{ida}",
                        instance.id,
                    )
            else:
                raise ValueError("Expected relationship config to be of type RelationshipConfigEnum")
            pipeline.execute()


def one_to_many(
    foreign_type: Union[str, Type[U]],
    backref: Optional[str] = None,
    lazy=True,
):
    return Relationship(
        foreign_type=foreign_type,
        config=RelationshipConfigEnum.ONE_TO_MANY,
        backref=backref,
        lazy=lazy,
    )


def one_to_one(
    foreign_type: Union[str, Type[U]],
    backref: Optional[str] = None,
    lazy=True,
):
    return Relationship(
        foreign_type=foreign_type,
        config=RelationshipConfigEnum.ONE_TO_ONE,
        backref=backref,
        lazy=lazy,
    )


def many_to_one(
    foreign_type: Union[str, Type[U]],
    backref: Optional[str] = None,
    lazy=True,
):
    return Relationship(
        foreign_type=foreign_type,
        config=RelationshipConfigEnum.MANY_TO_ONE,
        backref=backref,
        lazy=lazy,
    )


def many_to_many(
    foreign_type: Union[str, Type[U]],
    backref: Optional[str] = None,
    lazy=True,
):
    return Relationship(
        foreign_type=foreign_type,
        config=RelationshipConfigEnum.MANY_TO_MANY,
        backref=backref,
        lazy=lazy,
    )



from typing import (
    List,
    Type,
    TypeVar,
    Set,
    ClassVar,
    Iterable,
    Callable,
    Optional,
)
from dataclasses import dataclass, field, fields, Field
from uuid import uuid4
from collections import OrderedDict
import json
from dataclasses_jsonschema import JsonSchemaMixin
from redis.lock import Lock

from redorm.exceptions import (
    InstanceNotFound,
    UniqueContstraintViolation,
    UnknownFieldName,
    FilterOnUnindexedField,
    MultipleInstancesReturned,
)
from redorm.client import red

S = TypeVar("S", bound="RedormBase")

all_models = dict()


class Query:
    pipeline_results: Optional[List]

    def __init__(self):
        self.results = []
        self.pipeline_results = None
        self.resolvers: List[Callable[["Query"], None]] = []
        self.pipeline = red.client.pipeline()

    def execute(self) -> List:
        self.pipeline_results: List = self.pipeline.execute()
        results = []
        while self.resolvers:
            resolver = self.resolvers.pop()
            try:
                results.append(resolver(self))
            except InstanceNotFound:
                pass
        return list(reversed(results))


@dataclass
class RedormBase(JsonSchemaMixin):
    id: str = field(metadata={"unique": True})
    _relationships: ClassVar = OrderedDict()

    @classmethod
    def get(cls: Type[S], instance_id=None, **kwargs) -> S:
        if instance_id is not None and len(kwargs) == 0:
            query = Query()
            cls._get(query, instance_id)
            res = query.execute()
            if res:
                return res[0]
            else:
                raise InstanceNotFound
        else:
            if instance_id is None:
                if kwargs:
                    instance_ids = cls._list_ids(**kwargs)
                else:
                    raise InstanceNotFound
            else:
                instance_ids = cls._list_ids(id=instance_id, **kwargs)

            if len(instance_ids) == 1:
                instance_id = instance_ids.pop()
                query = Query()
                cls._get(query, instance_id)
                return query.execute()[0]
            elif len(instance_ids) > 1:
                raise MultipleInstancesReturned
            else:
                raise InstanceNotFound

    @classmethod
    def _resolve(cls: Type[S], query: Query) -> S:
        for rel_name, relation in reversed(cls._relationships.items()):
            relation.cached_ref = query.pipeline_results.pop()
            if not relation.lazy:
                res = query.pipeline_results.pop()
                if res is None:
                    relation.cached_class = None
                elif isinstance(res, list):
                    relation.cached_class = [
                        relation.get_foreign_type().from_json(d, validate=False) for d in res if d is not None
                    ]
                else:
                    relation.cached_class = relation.get_foreign_type().from_json(res, validate=False)
        data = query.pipeline_results.pop()
        if data is None:
            print(
                f"None for data, class={cls.__name__!r}, query.pipeline_results={query.pipeline_results!r}, query.resolvers={query.resolvers!r}"
            )
            raise InstanceNotFound
        return cls.from_json(data, validate=False)

    @classmethod
    def _get(cls: Type[S], query: Query, instance_id: str):
        query.pipeline.get(f"{cls.__name__}:member:{instance_id}")
        for rel_name, relation in cls._relationships.items():
            rel_key = f"{cls.__name__}:relationship:{rel_name}:{instance_id}"
            if not relation.lazy:
                if relation.to_many:
                    query.pipeline.evalsha(
                        red.get_set_indirect_sha,
                        2,
                        f"{relation.get_foreign_type().__name__}:member:",
                        rel_key,
                    )
                else:
                    query.pipeline.evalsha(
                        red.get_key_indirect_sha,
                        1,
                        rel_key,
                    )
            if relation.to_many:
                query.pipeline.smembers(rel_key)
            else:
                query.pipeline.get(rel_key)

        query.resolvers.append(cls._resolve)

    @classmethod
    def get_bulk(cls: Type[S], instance_ids: Set[str]) -> Iterable[S]:
        if len(instance_ids) == 0:
            return []
        query = Query()
        for instance_id in instance_ids:
            cls._get(query, instance_id)
        return query.execute()

    @classmethod
    def create(cls: Type[S], **kwargs) -> S:
        # Handle non-relationship parts
        instance_fields = fields(cls)
        field_values = {f.name: kwargs.get(f.name) for f in instance_fields if f.name in kwargs}
        new_id = str(uuid4())
        new_instance = cls.from_dict(dict(id=new_id, **field_values))
        for k, v in kwargs.items():
            if k not in field_values:
                setattr(new_instance, k, v)
        new_instance.save()
        return new_instance

    @classmethod
    def _list_ids(cls, **kwargs) -> Set[str]:
        field_dict = {f.name: f for f in fields(cls)}
        pre_pipeline = red.client.pipeline()
        indexes = set()
        if not kwargs:
            return set()
        try:
            for k, v in kwargs.items():
                if k in field_dict:
                    f: Field = field_dict[k]
                    if field_dict[k].metadata.get("unique"):
                        if v is None:
                            pre_pipeline.smembers(f"{cls.__name__}:keynull:{f.name}")
                        else:
                            pre_pipeline.hget(
                                f"{cls.__name__}:key:{k}",
                                cls._encode_field(f.type, v, omit_none=False),
                            )
                    elif field_dict[k].metadata.get("index"):
                        if v is None:
                            indexes.add(f"{cls.__name__}:indexnull:{f.name}")
                        else:
                            indexes.add(
                                f"{cls.__name__}:index:{f.name}:{cls._encode_field(f.type, v, omit_none=False)}"
                            )
                    else:
                        raise FilterOnUnindexedField(f"Trying to filter on unindexed field: {k}")
                elif isinstance(cls.__dict__[k], IRelationship):
                    rel = cls.__dict__[k]
                    if rel.to_many:
                        raise NotImplementedError("Can't filter based off to-many relationships")
                    if rel.backref is None:
                        raise NotImplementedError("Filtering on a relationship requires a backref")
                    rel_type = rel.get_foreign_type().__name__
                    if not isinstance(v, str):
                        v = v.id
                    ref = f"{rel_type}:relationship:{rel.backref}:{v}"
                    if rel.many_to:
                        indexes.add(ref)
                    else:
                        pre_pipeline.get(ref)

        except KeyError as e:
            raise UnknownFieldName(*e.args) from e
        if indexes:
            pre_pipeline.sinter(indexes)
        results = pre_pipeline.execute()
        sets = [r if isinstance(r, set) else {r} for r in results if r is not None]
        if not sets:
            return set()
        ret = sets[0]
        for s in sets[1:]:
            ret.intersection_update(s)
        return ret

    @classmethod
    def list(cls: Type[S], **kwargs) -> Iterable[S]:
        if len(kwargs) > 0:
            member_ids = cls._list_ids(**kwargs)
        else:
            member_ids = red.client.smembers(f"{cls.__name__}:all")
        return cls.get_bulk(member_ids)

    def lock(
        self,
        timeout=None,
        sleep=0.1,
        blocking_timeout=None,
        lock_class=None,
        thread_local=True,
    ) -> Lock:
        return red.client.lock(
            f"{self.__class__.__name__}:userlock:{self.id}",
            timeout=timeout,
            sleep=sleep,
            lock_class=lock_class,
            blocking_timeout=blocking_timeout,
            thread_local=thread_local,
        )

    def delete(self):
        # TODO: Handle removal of relationships
        instance_id = self.id
        cls_name = self.__class__.__name__
        with red.client.lock(f"{cls_name}:lock:{instance_id}"):
            self.refresh()
            instance_dict = self.to_dict(omit_none=True)
            p = red.client.pipeline()
            p.srem(f"{cls_name}:all", instance_id)
            p.unlink(f"{cls_name}:member:{instance_id}")
            for f in fields(self.__class__):
                val = instance_dict.get(f.name)
                if f.metadata.get("unique"):
                    if val is None:
                        p.srem(f"{cls_name}:keynull:{f.name}", instance_id)
                    else:
                        p.hdel(f"{cls_name}:key:{f.name}", val)
                elif f.metadata.get("index"):
                    if val is None:
                        p.srem(f"{cls_name}:indexnull:{f.name}", instance_id)
                    else:
                        p.srem(
                            f"{cls_name}:index:{f.name}:{self.__class__._encode_field(f.type, instance_dict[f.name], omit_none=True)}",
                            self.id,
                        )
            p.execute(raise_on_error=True)

    def refresh(self) -> None:
        latest = red.client.get(f"{self.__class__.__name__}:member:{self.id}")
        if latest is None:
            raise InstanceNotFound
        for k, v in json.loads(latest).items():
            setattr(self, k, v)
        for k, v in self._relationships.items():
            v.cached_class = None
            v.cached_ref = None

    def save(self) -> None:
        old_json = red.client.get(f"{self.__class__.__name__}:member:{self.id}")
        if old_json is None:
            old_dict = {}
            new = True
        else:
            old = self.__class__.from_json(old_json)
            old_dict = old.to_dict(omit_none=False)
            new = False
        instance_dict = self.to_dict(omit_none=True)
        instance_fields = fields(self.__class__)
        cls_name = self.__class__.__name__

        # Pipeline to rollback changes on error
        revert_pipeline = red.client.pipeline()

        # Pipeline to ensure uniqueness of unique fields
        unique_pipeline = red.client.pipeline()
        unique_pipeline.sadd(f"{self.__class__.__name__}:all", self.id)
        # TODO: Handle rollback
        changed_unique_fields_non_null = [
            f
            for f in instance_fields
            if f.metadata.get("unique")
            and instance_dict.get(f.name) is not None
            and (instance_dict.get(f.name) != old_dict.get(f.name) or new)
        ]
        changed_unique_fields_null = [
            f
            for f in instance_fields
            if f.metadata.get("unique")
            and instance_dict.get(f.name) is None
            and (instance_dict.get(f.name) != old_dict.get(f.name) or new)
        ]
        for f in changed_unique_fields_non_null:
            unique_pipeline.hget(f"{cls_name}:key:{f.name}", instance_dict[f.name])
            unique_pipeline.hsetnx(f"{cls_name}:key:{f.name}", instance_dict[f.name], self.id)
            if old_dict.get(f.name) is not None:
                revert_pipeline.hset(f"{cls_name}:key:{f.name}", old_dict[f.name], self.id)
                unique_pipeline.ping()
            else:
                revert_pipeline.sadd(f"{cls_name}:keynull:{f.name}", self.id)
                unique_pipeline.srem(f"{cls_name}:keynull:{f.name}", self.id)

        for f in changed_unique_fields_null:
            unique_pipeline.sadd(f"{cls_name}:keynull:{f.name}", self.id)
            revert_pipeline.srem(f"{cls_name}:keynull:{f.name}", self.id)
        unique_pipeline_result_all = unique_pipeline.execute()
        unique_pipeline_result = unique_pipeline_result_all[1:]
        if unique_pipeline_result_all[0] == 1:
            revert_pipeline.srem(f"{self.__class__.__name__}:all", self.id)
        unique_violating_fields = []
        for i, r in enumerate(unique_pipeline_result[: len(changed_unique_fields_non_null) * 3 : 3]):
            if r is None:
                revert_pipeline.hdel(
                    f"{cls_name}:key:{changed_unique_fields_non_null[i].name}",
                    instance_dict[changed_unique_fields_non_null[i].name],
                )
            elif r != self.id:
                revert_pipeline.hdel(
                    f"{cls_name}:key:{changed_unique_fields_non_null[i].name}",
                    instance_dict[changed_unique_fields_non_null[i].name],
                )
                # Failed to set as key already exists
                unique_violating_fields.append(changed_unique_fields_non_null[i].name)

        try:
            if len(unique_violating_fields) > 0:
                raise UniqueContstraintViolation(f"New object contained non-unique values: {unique_violating_fields!r}")
            # Fields that require indexing and are not indexed by uniqueness
            # Index is stored as a set of ids
            indexed_non_unique_fields = [
                f
                for f in instance_fields
                if f.metadata.get("index")
                and not f.metadata.get("unique")
                and (instance_dict.get(f.name) != old_dict.get(f.name) or new)
            ]
            # Pipeline to index indexable fields
            index_pipeline = red.client.pipeline()
            for f in indexed_non_unique_fields:
                if instance_dict.get(f.name) is None:
                    index_pipeline.sadd(
                        f"{cls_name}:indexnull:{f.name}",
                        self.id,
                    )
                    if new:
                        index_pipeline.ping()
                    else:
                        index_pipeline.srem(
                            f"{cls_name}:index:{f.name}:{self.__class__._encode_field(f.type, old_dict[f.name], omit_none=True)}",
                            self.id,
                        )
                        revert_pipeline.sadd(
                            f"{cls_name}:index:{f.name}:{self.__class__._encode_field(f.type, old_dict[f.name], omit_none=True)}",
                            self.id,
                        )
                else:
                    index_pipeline.sadd(
                        f"{cls_name}:index:{f.name}:{self.__class__._encode_field(f.type, instance_dict[f.name], omit_none=True)}",
                        self.id,
                    )
                    if new:
                        index_pipeline.ping()
                    elif old_dict.get(f.name) is None:
                        index_pipeline.srem(
                            f"{cls_name}:indexnull:{f.name}",
                            self.id,
                        )
                        revert_pipeline.sadd(f"{cls_name}:indexnull:{f.name}", self.id)
                    else:
                        index_pipeline.srem(
                            f"{cls_name}:index:{f.name}:{self.__class__._encode_field(f.type, old_dict[f.name], omit_none=True)}",
                            self.id,
                        )
                        revert_pipeline.sadd(
                            f"{cls_name}:index:{f.name}:{self.__class__._encode_field(f.type, old_dict[f.name], omit_none=True)}",
                            self.id,
                        )

            index_pipeline_results = index_pipeline.execute()
            for i, r in enumerate(index_pipeline_results[::2]):
                if r == 1:
                    f = indexed_non_unique_fields[i]
                    if instance_dict.get(f.name) is None:
                        revert_pipeline.srem(
                            f"{cls_name}:indexnull:{f.name}",
                            self.id,
                        )
                    else:
                        revert_pipeline.srem(
                            f"{cls_name}:index:{f.name}:{self.__class__._encode_field(f.type, instance_dict[f.name], omit_none=True)}",
                            self.id,
                        )

            pipe = red.client.pipeline()
            pipe.set(
                f"{cls_name}:member:{self.id}",
                json.dumps(instance_dict, sort_keys=True),
            )
            pipe.execute()
        except Exception as e:
            revert_pipeline.execute()
            raise e

    def update(self, **kwargs):
        with red.client.lock(f"{self.__class__.__name__}:lock:{self.id}"):
            self.refresh()
            for k, v in kwargs.items():
                setattr(self, k, v)
            self.save()

    def __init_subclass__(cls, **kwargs):
        all_models[cls.__name__] = cls
        super().__init_subclass__(**kwargs)

    def __hash__(self):
        return hash(f"{self.__class__.__name__}:{self.id}")


class IRelationship:
    pass



class RedormException(Exception):
    pass


class InstanceNotFound(RedormException):
    pass


class DevelopmentServerUsedInProd(RedormException):
    pass


class UniqueContstraintViolation(RedormException):
    pass


class UnknownFieldName(RedormException):
    pass


class FilterOnUnindexedField(RedormException):
    pass


class MultipleInstancesReturned(RedormException):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['redorm']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses-jsonschema[fast-validation]>=2.13.0,<3.0.0',
 'environs>=9.3.0,<10.0.0',
 'fakeredis>=1.4.5,<2.0.0',
 'redis>=3.5.3,<4.0.0']

setup_kwargs = {
    'name': 'redorm',
    'version': '0.6.10',
    'description': 'A simple redis ORM',
    'long_description': '# RedORM\n\nA simple Redis ORM that only a madman would use in production.  \nThe red in RedORM both means Redis as well as the color red, as red is the fastest colour!\n\n## Quick Start\n\nTo install `pip install redorm`\n\n```python\nfrom dataclasses import dataclass\nfrom redorm import RedormBase, one_to_one, one_to_many, many_to_one, many_to_many\n\n\n@dataclass\nclass Person(RedormBase):\n    name: str\n    age: int\n    siblings = many_to_many(foreign_type="Person", backref="siblings")\n    dad = many_to_one(foreign_type="Person", backref="children")\n    children = one_to_many(foreign_type="Person", backref="dad")\n    favourite_color = one_to_one("Color", backref="liker")\n\n\n@dataclass\nclass Color(RedormBase):\n    name: str\n    liker = one_to_one(Person, backref="favourite_color")\n```\n```python\n>>> red = Color.create(name="Red")\n>>> homer = Person.create(name="Homer", age=50, favourite_color=red)\n>>> print(repr(homer.favourite_color))\nColor(id=\'dcb9aa50-554a-40a5-9acf-7d86c982e5ee\', name=\'Red\')\n>>> print(repr(homer.children))\n[]\n>>> bart = Person.create(name="Bart", age=11, dad=homer)\n>>> print(repr(homer.children))\n[Person(id=\'424cd574-5382-4d34-89da-7233b3928405\', name=\'Bart\', age=11)]\n>>> print(repr(bart.favourite_color))\nNone\n>>> blue = Color.create(name="Blue", liker=bart)\n>>> print(repr(bart.favourite_color))\nColor(id=\'dc9df3c2-c592-4d87-a45e-f88a346342b4\', name=\'Blue\')\n>>> print(repr(blue.liker))\nPerson(id=\'424cd574-5382-4d34-89da-7233b3928405\', name=\'Bart\', age=11)\n>>> lisa = Person.create(name="Lisa", age=9, dad=homer.id, siblings=[bart])\n>>> print(repr(homer.children))\n[Person(id=\'205a459a-572c-41af-bae3-e6e730aada97\', name=\'Lisa\', age=9), Person(id=\'424cd574-5382-4d34-89da-7233b3928405\', name=\'Bart\', age=11)]\n>>> bart.dad = None\n>>> print(repr(homer.children))\n[Person(id=\'205a459a-572c-41af-bae3-e6e730aada97\', name=\'Lisa\', age=9)]\n```\n\n## Why Redorm?\n\n- Thread Safe\n- Very fast\n- Super simple to use\n- Very little boilerplate\n\n# Why not Redorm?\n\n- Made in an afternoon\n- Unlikely to be maintained\n- Not thoroughly tested\n- NOT THOROUGHLY TESTED\n- Writing your own ORM is fantastic for learning, but you should not use it in prod\n',
    'author': 'Jack Adamson',
    'author_email': 'jack@mrfluffybunny.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
